/*
 * Copyright 2023 David A. Holland. All Rights Reserved.
 *
 * This software is provided 'as-is', without any express or implied
 * warranty.  In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 */

open Tree

module StringMap = Map.Make(String)

let rec eval'expr env e0 =
   match e0 with
   | CONST k -> k
   | READVAR x -> StringMap.find x env
   | NEG e1 ->
        let v1 = eval'expr env e1 in
        -. v1
   | ADD (e1, e2) ->
        let v1 = eval'expr env e1 in
        let v2 = eval'expr env e2 in
        v1 +. v2
   | SUB (e1, e2) ->
        let v1 = eval'expr env e1 in
        let v2 = eval'expr env e2 in
        v1 -. v2
   | MUL (e1, e2) ->
        let v1 = eval'expr env e1 in
        let v2 = eval'expr env e2 in
        v1 *. v2
   | DIV (e1, e2) ->
        let v1 = eval'expr env e1 in
        let v2 = eval'expr env e2 in
        v1 /. v2

let eval'stmt env s0 =
   match s0 with
   | COMPUTE e ->
        let v = eval'expr env e in
        print_string (string_of_float v ^ "\n");
        env
   | ASSIGN (x, e) ->
        let v = eval'expr env e in
        print_string (x ^ " = " ^ string_of_float v ^ "\n");
        StringMap.add x v env

let rec repl_once env =
   print_string "> ";
   flush stdout;
   let line = read_line () in
   if line = ":q" then
      None
   else begin
      try
         let t = Lexer.read line in
         let env' = eval'stmt env t in
         Some env'
      with e ->
         prerr_string ("Exception: " ^ Printexc.to_string e ^ "\n");
         flush stderr;
         Some env
   end

let rec repl env =
   match repl_once env with
   | None -> ()
   | Some env' -> repl env'

let _ = repl StringMap.empty
